package org.eclipse.team.examples.pessimistic.ui;

/*
 * (c) Copyright IBM Corp. 2000, 2001.
 * All Rights Reserved.
 */

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.ILabelDecorator;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.LabelProviderChangedEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Display;
import org.eclipse.team.core.RepositoryProvider;
import org.eclipse.team.core.TeamPlugin;
import org.eclipse.team.examples.pessimistic.PessimisticProvider;
import org.eclipse.team.ui.ISharedImages;
import org.eclipse.team.ui.TeamUIPlugin;

/*
 * A simple example decorator.
 */
public class PessimisticDecorator extends LabelProvider implements ILabelDecorator, IResourceChangeListener {

	// Used to exit the isDirty resource visitor
	private static final CoreException CORE_EXCEPTION = new CoreException(new Status(IStatus.OK, "id", 1, "", null));

	public PessimisticDecorator() {
		ResourcesPlugin.getWorkspace().addResourceChangeListener(this);
	}
		
	/*
	 * @see ITeamDecorator#getText(String, IResource)
	 */
	public String decorateText(String text, Object element) {
		IResource resource = getResource(element);
		if(resource==null || resource.getType() != IResource.FILE)
			return text;
		PessimisticProvider provider= (PessimisticProvider)RepositoryProvider.getProvider(resource.getProject(), PessimisticProvider.ID);
		if (provider == null) {
			return text;
		}
		if (provider.isCheckedout(resource)) {
			return "> " + text;
		} else {
			return text;
		}
	}

	/*
	 * @see ITeamDecorator#getImage(IResource)
	 */
	public Image decorateImage(Image image, Object element) {
//		IResource resource = getResource(element);
//		if(resource==null || resource.getType() == IResource.ROOT) return image;
//		SimpleProvider p = getSimpleProvider(resource);
//		if(p==null) return image;
//		
//		List overlays = new ArrayList(5);	
//		
//		if(isDirty(resource)) {
//			overlays.add(TeamUIPlugin.getPlugin().getImageDescriptor(ISharedImages.IMG_DIRTY_OVR));
//		}
//		
//		if(p!=null && p.hasRemote(resource)) {
//			overlays.add(TeamUIPlugin.getPlugin().getImageDescriptor(ISharedImages.IMG_CHECKEDIN_OVR));
//		} else if(p!=null && p.isCheckedOut(resource)) {
//			overlays.add(TeamUIPlugin.getPlugin().getImageDescriptor(ISharedImages.IMG_CHECKEDOUT_OVR));
//		}
//		return new OverlayIcon(image.getImageData(), new ImageDescriptor[][] {(ImageDescriptor[])overlays.toArray(new ImageDescriptor[overlays.size()])}, new Point(16, 16)).createImage();
		return image;
	}
		
	/*
	 * @see IResourceChangeListener#resourceChanged(IResourceChangeEvent)
	 */
	public void resourceChanged(IResourceChangeEvent event) {		
		processChanged(event.getDelta());
	}

	/*
	 * @see IResourceStateChangeListener#resourceStateChanged(IResource[])
	 */
	public void resourceStateChanged(IResource[] changedResources) {
		for (int i = 0; i < changedResources.length; i++) {
			processStateChanged(changedResources[i]);			
		}
	}
	
	private void processStateChanged(IResource resource) {
		IResource current = resource;
		List events = new ArrayList();
		while (current.getType() != IResource.ROOT) {
			events.add(new LabelProviderChangedEvent(this,current));
			current = current.getParent();
		}
		postLabelEvents((LabelProviderChangedEvent[]) events.toArray(new LabelProviderChangedEvent[events.size()]));
	}
	
	/**
	 * Generate label change events for each resource in the change delta. This will
	 * force a refresh for all decorators associated with the resources in the change
	 * event.
	 */
	private void processChanged(IResourceDelta delta) {
		final LinkedList events = new LinkedList();
		try {
			delta.accept(new IResourceDeltaVisitor() {
				public boolean visit(IResourceDelta delta) throws CoreException {
					IResource resource = delta.getResource();
					int type = resource.getType();

					// skip workspace root
					if (type == IResource.ROOT) {
						return true;
					}
					
					// if this is a change event for a resource not associated with a simple
					// project, then stop processing the deltas and return an empty event list.
					PessimisticProvider p = getSimpleProvider(resource);
					if(p==null) return false;

					// don't care about deletions
					if (delta.getKind() == IResourceDelta.REMOVED) {
						return false;
					}

//					// ignore subtrees that don't have remotes
//					if (!p.hasRemote(resource)) {
//						return false;
//					}
					// chances are the team outgoing bit needs to be updated
					// if any child has changed.
					events.addFirst(new LabelProviderChangedEvent(PessimisticDecorator.this, resource));
					return true;
				}
			});
		} catch (CoreException e) {
			TeamUIPlugin.log(e.getStatus());
		}
		// post label events for resource delta, this will force a refresh of the decorations
		// for the elements in the delta.
		postLabelEvents((LabelProviderChangedEvent[]) events.toArray(new LabelProviderChangedEvent[events.size()]));
	}
	
	/**
	 * Returns the simple team provider associate with the resource or <code>null</code> if
	 * the <code>SimpleTeamProvider</code> is not associated with that resource's project.
	 */
	private PessimisticProvider getSimpleProvider(IResource resource) {
		try {
			return (PessimisticProvider) RepositoryProvider.getProvider(resource.getProject(), PessimisticProvider.ID);			
		} catch(ClassCastException e) {
			return null;
		}
	}

	/**
	 * Returns the resource for the given input object, or null if there is no resource 
	 * associated with it.
	 */
	private IResource getResource(Object object) {
		if (object instanceof IResource) {
			return (IResource)object;
		}
		if (object instanceof IAdaptable) {
			return (IResource)((IAdaptable)object).getAdapter(IResource.class);
		}
		return null;
	}

	/**
	 * Post the label events to the UI thread
	 */
	private void postLabelEvents(final LabelProviderChangedEvent[] events) {
		if (events.length > 0) {
			Display.getDefault().asyncExec(new Runnable() {
				public void run() {
					for (int i = 0; i < events.length; i++) {
						fireLabelProviderChanged(events[i]);
					}
				}
			});
		}
	} 
	
	/**
	 * Consider a resource dirty if any of its members are dirty or has no
	 * corresponding remote resource (e.g. has never been checked in).
	 */
	private boolean isDirty(IResource resource) {
//		try {
//			resource.accept(new IResourceVisitor() {
//				public boolean visit(IResource resource) throws CoreException {
//					RepositoryProvider p = RepositoryProvider.getProvider(resource.getProject());
//					if(p==null) {
//						return false;
//					}
//					// project are never dirty - continue looking at children
//					if(resource.getType() == IResource.PROJECT) {
//						return true;
//					}
//					if( ((SimpleProvider)p).isDirty(resource) ) {
//						throw CORE_EXCEPTION;
//					} else if(!((SimpleProvider)p).hasRemote(resource)) {
//						throw CORE_EXCEPTION;
//					} else {
//						return true;
//					}
//				}
//			}, IResource.DEPTH_INFINITE, true);
//		} catch (CoreException e) {
//			//if our exception was caught, we know there's a dirty child
//			return e == CORE_EXCEPTION;
//		}
		return false;
	}
	/**
	 * @see IBaseLabelProvider#dispose()
	 */
	public void dispose() {
		ResourcesPlugin.getWorkspace().removeResourceChangeListener(this);
		super.dispose();
	}

}